/** Copyright (C) 2017 by Bundesanstalt für Gewässerkunde
 * Software engineering by
 *  Björnsen Beratende Ingenieure GmbH
 *  Dr. Schumacher Ingenieurbüro für Wasser und Umwelt
 *
 * This file is Free Software under the GNU AGPL (>=v3)
 * and comes with ABSOLUTELY NO WARRANTY! Check out the
 * documentation coming with Dive4Elements River for details.
 */
package org.dive4elements.river.artifacts.uinfo.salix;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableMap;

import org.dive4elements.river.artifacts.WINFOArtifact;
import org.dive4elements.river.artifacts.access.ComputationRangeAccess;
import org.dive4elements.river.artifacts.common.AbstractResultType;
import org.dive4elements.river.artifacts.common.GeneralResultType;
import org.dive4elements.river.artifacts.common.ResultRow;
import org.dive4elements.river.artifacts.model.Calculation;
import org.dive4elements.river.artifacts.model.WstValueTable;
import org.dive4elements.river.artifacts.model.WstValueTable.QPosition;
import org.dive4elements.river.artifacts.model.WstValueTableFactory;
import org.dive4elements.river.artifacts.sinfo.common.RiverInfoProvider;
import org.dive4elements.river.artifacts.sinfo.tkhstate.WinfoArtifactWrapper;
import org.dive4elements.river.artifacts.uinfo.UINFOArtifact;
import org.dive4elements.river.artifacts.uinfo.common.UInfoResultType;
import org.dive4elements.river.artifacts.uinfo.salix.SalixLineAccess.ScenarioType;
import org.dive4elements.river.artifacts.uinfo.vegetationzones.VegetationZoneServerClientXChange;
import org.dive4elements.river.model.Gauge;
import org.dive4elements.river.model.MainValue;
import org.dive4elements.river.model.MainValueType.MainValueTypeKey;
import org.dive4elements.river.utils.Formatter;

/**
 * Calculation of the result rows of the u-info salix line calc mode
 *
 * @author Matthias Schäfer
 */
final class SalixLineCalculator {

    private static final BigDecimal SALIX_DISTANCE = new BigDecimal("2.31");
    private final List<ResultRow> rows = new ArrayList<>();

    private final RiverInfoProvider riverInfoProvider;

    private final Map<Gauge, QPosition> gaugeMwPos;
    private final Map<Gauge, QPosition> gaugeMnwPos;
    private final Map<Gauge, QPosition> gaugeMhwPos;
    private final Map<Gauge, QPosition> gaugeHw5Pos;
    private QPosition refGaugeMwPos;
    private QPosition refGaugeMnwPos;
    private QPosition refGaugeMhwPos;
    private QPosition refGaugeHw5Pos;
    private Gauge firstGauge;

    private Calculation problems;

    private WstValueTable wst;

    public SalixLineCalculator(final RiverInfoProvider riverInfoProvider) {
        this.riverInfoProvider = riverInfoProvider;
        this.gaugeMwPos = new HashMap<>();
        this.gaugeMnwPos = new HashMap<>();
        this.gaugeMhwPos = new HashMap<>();
        this.gaugeHw5Pos = new HashMap<>();
    }

    /**
     * Calculate the salix line result rows
     */
    public void execute(final Calculation problems, final UINFOArtifact uinfo, final NavigableMap<Double, List<Double>> rangeScenarios,
            final ScenarioType scenarioType, final String[] scenarioLabels, final String rangeString, final String additionalString,
            final SalixLineCalculationResults results) {

        this.problems = problems;
        this.wst = WstValueTableFactory.getTable(this.riverInfoProvider.getRiver());

        fetchGaugeMainValuePositions();

        final WINFOArtifact winfo = new WinfoArtifactWrapper(uinfo);
        winfo.addStringData("ld_mode", "distance");
        winfo.addStringData("ld_step", "100");
        for (final double station : new ComputationRangeAccess(winfo).getKms()) {
            this.rows.add(createRow(station, rangeScenarios));
        }
        if (scenarioType == ScenarioType.REGIONAL)
            results.addResult(new SalixLineCalculationRegionalResult("Salix-regional", scenarioLabels, rangeString, additionalString, this.rows), problems);
        else if (scenarioType == ScenarioType.SUPRAREGIONAL)
            results.addResult(new SalixLineCalculationSupraRegionalResult("Salix-supra", scenarioLabels, rangeString, additionalString, this.rows), problems);
        else if (scenarioType == ScenarioType.HISTORICAL)
            results.addResult(new SalixLineCalculationHistoricalResult("Salix-hist", scenarioLabels, rangeString, additionalString, this.rows), problems);
        else
            results.addResult(new SalixLineCalculationResult("Salix-simple", this.rows), problems);
    }

    /**
     * Fetch MQ, MNQ and MHQ of all gauges and determine the wst QPosition for each one
     */
    private void fetchGaugeMainValuePositions() {
        this.gaugeMwPos.clear();
        this.gaugeMnwPos.clear();
        this.gaugeMhwPos.clear();
        this.gaugeHw5Pos.clear();

        this.firstGauge = null;
        for (final Gauge gauge : this.riverInfoProvider.getGauges()) {
            this.gaugeMwPos.put(gauge, null);
            this.gaugeMnwPos.put(gauge, null);
            this.gaugeMhwPos.put(gauge, null);
            this.gaugeHw5Pos.put(gauge, null);
            final double gaugeKm = gauge.getStation().doubleValue();
            for (final MainValue mv : MainValue.getValuesOfGaugeAndType(gauge, MainValueTypeKey.Q)) {
                if (mv.getMainValue().getName().equalsIgnoreCase("mq"))
                    this.gaugeMwPos.put(gauge, this.wst.getQPosition(gaugeKm, mv.getValue().doubleValue()));
                else if (mv.getMainValue().getName().equalsIgnoreCase("mnq"))
                    this.gaugeMnwPos.put(gauge, this.wst.getQPosition(gaugeKm, mv.getValue().doubleValue()));
                else if (mv.getMainValue().getName().equalsIgnoreCase("mhq"))
                    this.gaugeMhwPos.put(gauge, this.wst.getQPosition(gaugeKm, mv.getValue().doubleValue()));
                else if (mv.getMainValue().getName().equalsIgnoreCase("hq5"))
                    this.gaugeHw5Pos.put(gauge, this.wst.getQPosition(gaugeKm, mv.getValue().doubleValue()));
            }
            if (this.firstGauge == null) {
                this.refGaugeMwPos = this.gaugeMwPos.get(gauge);
                this.refGaugeMnwPos = this.gaugeMnwPos.get(gauge);
                this.refGaugeMhwPos = this.gaugeMhwPos.get(gauge);
                this.refGaugeHw5Pos = this.gaugeHw5Pos.get(gauge);
                this.firstGauge = gauge;
            }
        }
        if (this.refGaugeMwPos == null)
            this.problems.addProblem("uinfo_salix_calc.warning.missing_mq");
        else {
            if (this.refGaugeMhwPos == null)
                this.problems.addProblem("uinfo_salix_calc.warning.missing_mhq");
            if (this.refGaugeMnwPos == null)
                this.problems.addProblem("uinfo_salix_calc.warning.missing_mnq");
        }
    }

    /**
     * Create a result row for a station and its gauge, and add w-q-values as selected
     */
    private ResultRow createRow(final double station, final NavigableMap<Double, List<Double>> rangeScenarios) {

        final ResultRow row = ResultRow.create();
        row.putValue(GeneralResultType.station, station);
        // Find station's gauge (obsolete version which calculates gauge-wise)
        // final Gauge gauge = this.riverInfoProvider.getGauge(station, true);
        // Interpolate mnw, mw, and mhw
        // final double mnw = interpolateW(station, this.gaugeMnwPos.get(gauge));
        // final double mw = interpolateW(station, this.gaugeMwPos.get(gauge));
        // final double mhw = interpolateW(station, this.gaugeMhwPos.get(gauge));
        final double mnw = interpolateW(station, this.refGaugeMnwPos);
        final double mw = interpolateW(station, this.refGaugeMwPos);
        final double mhw = interpolateW(station, this.refGaugeMhwPos);
        final double hw5 = interpolateW(station, this.refGaugeHw5Pos);
        row.putValue(UInfoResultType.waterlevelMNW, mnw);
        row.putValue(UInfoResultType.waterlevelMW, mw);
        row.putValue(UInfoResultType.waterlevelMHW, mhw);
        row.putValue(UInfoResultType.waterlevelMH5, hw5);

        // Calc salix-line and mw-mnw
        row.putValue(UInfoResultType.salixline, calcSalix(mhw, mw, 0.0));
        row.putValue(UInfoResultType.salix_mw_mnw, calcMwmnw(mw, mnw));
        final double salixw = Formatter.roundW(mhw).subtract(SALIX_DISTANCE).doubleValue();
        row.putValue(UInfoResultType.salixw, salixw);
        // Calc scenario values (always all scenario types set, Result variant extracts the fields needed)
        final List<SalixScenario> scenarios = new ArrayList<>();
        final List<Double> deltaws = getDeltaWs(station, rangeScenarios);
        for (final Double deltaw : deltaws) {
            if (deltaw != null) {
                final double salix = calcSalix(mhw, mw, deltaw.doubleValue());
                final double scen = calcSalix(mhw, 0.0, deltaw.doubleValue());
                scenarios.add(new SalixScenario((int) (deltaw * 100), salix, scen));
            }
            else {
                scenarios.add(null);
            }
        }
        row.putValue(UInfoResultType.customMultiRowColSalixScenarios, scenarios);
        row.putValue(GeneralResultType.gaugeLabel, this.riverInfoProvider.findGauge(station));
        return row;
    }

    /**
     * Interpolates the W for a station with a fixed (virtual) wst column position
     */
    private double interpolateW(final double station, final QPosition qPosition) {
        if (qPosition != null)
            return this.wst.interpolateW(station, qPosition, this.problems);
        return Double.NaN;
    }

    /**
     * Calculates the salix value
     */
    private double calcSalix(final double mhw, final double mw, final double deltamw) {
        if (Double.isNaN(mw) || Double.isInfinite(mw) || Double.isNaN(mhw) || Double.isInfinite(mhw))
            return mhw - mw; // preserving NaN or Infinity
        return Formatter.roundW(mhw).subtract(SALIX_DISTANCE).subtract(Formatter.roundW(mw).add(Formatter.roundW(deltamw))).doubleValue();
    }

    /**
     * Calculates the inverse MW-MNW difference
     */
    private double calcMwmnw(final double mw, final double mnw) {
        if (Double.isNaN(mw) || Double.isInfinite(mw) || Double.isNaN(mnw) || Double.isInfinite(mnw))
            return mnw - mw; // preserving NaN or Inifinity
        return Formatter.roundW(mnw).subtract(Formatter.roundW(mw)).doubleValue();
    }

    /**
     * Gets the station-specific list of delta-ws of the active scenario, at least with one null item in any case
     */
    private List<Double> getDeltaWs(final double station, final NavigableMap<Double, List<Double>> rangeScenarios) {
        final Entry<Double, List<Double>> stationScenarios = rangeScenarios.floorEntry(station);
        if (stationScenarios != null) {
            return stationScenarios.getValue();
        }
        final List<Double> noScen = new ArrayList<>();
        noScen.add(null);
        return noScen;
    }

    /**
     * Find and return a height (iota, w main value) of a station in a previously calculated result
     */
    public double fetchStationHeight(final Calculation problems, final double station, final AbstractResultType resultType,
            final SalixLineCalculationResult result) {

        // Search the station in the previously calculated result rows
        final ResultRow stationRow = searchStation(station, result.getRows());
        if (stationRow != null)
            return stationRow.getDoubleValue(resultType);
        return Double.NaN;
    }

    /**
     * Searches the row of a station in a result rows collection
     */
    private ResultRow searchStation(final double station, final Collection<ResultRow> rows) {
        for (final ResultRow row : rows) {
            if (row.getDoubleValue(GeneralResultType.station) > station + 0.0001)
                return row;
        }
        return null;
    }

    /**
     * Computes the height of a vegetation zone limit for a station and a previously computed salix line result
     */
    public double computeVegetationZoneHeight(final Calculation problems, final double station, final int vegetationZoneType,
            final SalixLineCalculationResult result) {

        // Search the station in the previously calculated result rows
        final ResultRow stationRow = searchStation(station, result.getRows());
        if (stationRow == null)
            return Double.NaN;
        // Compute height from overflow duration days
        final List<VegetationZoneServerClientXChange> vzs = VegetationZoneServerClientXChange.getStandardList(null, null); // TODO river, context
        if ((vegetationZoneType >= 1) && (vegetationZoneType <= vzs.size())) {
            final int uefd = vzs.get(vegetationZoneType - 1).getMin_day_overflow();
            // Üfd = -70,559 ∗ ln((DGM - MW) + 0,5) + 80,711
            final double f1 = -70.559;
            final double f2 = -88.711;
            final double mw = stationRow.getDoubleValue(UInfoResultType.waterlevelMW);
            final double dgm = Math.exp((uefd - f2) / f1) + mw - 0.5;
            return dgm;
        }
        return Double.NaN;
    }
}
